module Config (Config (..), ConfigAction (..), getConfig, showHelp, showVersion) where

import System.Environment (getArgs, getProgName)
import Version

data ConfigAction = ActionShowHelp | ActionShowVersion | ActionExtractModels | ActionListModels | ActionShowGui
  deriving Show


data Config = Config {
    cfg_action         :: ConfigAction,
    cfg_quiet          :: Bool,
    cfg_inputFiles     :: [String],
    cfg_modelSelection :: Maybe [String]
    }


defaultConfig :: Config
defaultConfig = Config {
  cfg_action = ActionExtractModels,
  cfg_quiet = False,
  cfg_inputFiles = [],
  cfg_modelSelection = Nothing
  }

setShowHelp :: Config -> Config
setShowHelp cfg = cfg { cfg_action = ActionShowHelp }


data Param
  = P  [String] String (Config -> Config)
  | PX [String] String (Config -> Config)
  | PI [String] String (Config -> Int -> Config)
  | PS [String] String (Config -> String -> Config)


isPX :: Param -> Bool
isPX (PX _ _ _) = True
isPX _ = False

showAliasesAndDescription :: [String] -> String -> String 
showAliasesAndDescription aliases description
  | length aliasList <= 20 = (take 22 (aliasList ++ repeat ' ')) ++ description
  | otherwise = aliasList ++ "\n" ++ (replicate 22 ' ') ++ description
  where
    aliasList = go aliases

    go [x] = x
    go (x:xs) = x ++ " | " ++ go xs
    go [] = ""


instance Show Param where
  show (P  aliases description _) = showAliasesAndDescription aliases description
  show (PX aliases description _) = showAliasesAndDescription aliases description
  show (PI aliases description _) = showAliasesAndDescription aliases description
  show (PS aliases description _) = showAliasesAndDescription aliases description



programParams :: [Param]
programParams =
    [ PX ["-h", "/h", "-?", "/?", "-help", "/help", "--help"] "Shows this text." setShowHelp
    -- , P  ["--gui"] "Enables the graphical user interface" (\c -> c { cfg_action = ActionShowGui } )
    , PX ["-v"] "Print version." (\c -> c { cfg_action = ActionShowVersion } )
    , P  ["-q"] "Be lesser verbose." (\c -> c { cfg_quiet = True } )
    , P  ["-e", "--extract"] "Extract all or selected models from the given files." (\c -> c { cfg_action = ActionExtractModels } )
    , P  ["-l", "--list"] "List all models in the given files." (\c -> c { cfg_action = ActionListModels } )
    , PS ["-m", "--model"] "Model name to extract. Multiple models might be given."
        (\ c model -> c { cfg_modelSelection = case cfg_modelSelection  c of
            Nothing -> Just [model]
            (Just models) -> Just (model : models)
        } ) 
    ]



updateConfig :: Config -> [String] -> Config
updateConfig cfg ("":rest) = updateConfig cfg rest

updateConfig cfg ("--":rest) = updateConfig (cfg { cfg_inputFiles = (reverse rest) ++ (cfg_inputFiles cfg) }) []

updateConfig cfg (name:rest) = case findParam programParams of
    (Just (P  _ _ fn)) -> updateConfig (fn cfg) rest

    (Just (PX _ _ fn)) -> updateConfig (fn cfg) []
    
    (Just (PS _ _ fn)) -> case rest of
        (x:xs) -> updateConfig (fn cfg x) xs
        []     -> updateConfig (setShowHelp cfg) []

    (Just (PI _ _ fn)) -> case rest of 
        (x:xs) -> updateConfig (fn cfg (read x)) xs
        []     -> updateConfig (setShowHelp cfg) []

    Nothing -> updateConfig (cfg { cfg_inputFiles = name : (cfg_inputFiles cfg) }) rest

  where
    findParam (x@(P ps _ _):xs)
        | name `elem` ps = Just x

    findParam (x@(PX ps _ _):xs)
        | name `elem` ps = Just x

    findParam (x@(PI ps _ _):xs)
        | name `elem` ps = Just x

    findParam (x@(PS ps _ _):xs)
        | name `elem` ps = Just x

    findParam (_:xs) = findParam xs

    findParam [] = Nothing

updateConfig cfg [] = cfg { cfg_inputFiles = reverse (cfg_inputFiles cfg) }




getConfig :: IO Config
getConfig = getArgs >>= ( \ args -> pure $ case args of
    [] -> setShowHelp defaultConfig
    _  -> updateConfig defaultConfig args
    )



showVersion :: IO ()
showVersion = do
  name <- getProgName
  putStrLn $ name ++ " Version " ++ version ++ " (c) " ++ copyright
    

showHelp :: IO ()
showHelp = do
  showVersion
  putStrLn ""
  name <- getProgName
  putStrLn $ name ++ (parameterLine (filter (not . isPX) programParams)) ++ " [--] [inputFileName]"
  mapM_ (\ p -> putStrLn $ name ++ (parameterLine [p])) (filter isPX programParams)
  putStrLn ""
  mapM_ printLn programParams 
  putStrLn "--                    All parameter after this are interpreted as inputFileName."
  putStrLn ""
  putStrLn "inputFileName         PDBQT file with models to extract."
  where
    printLn x = print x >> putStrLn "" 

    parameterLine ((P aliases _ _):ps)  = " [" ++ go aliases ++ "]" ++ parameterLine ps
    parameterLine ((PX aliases _ _):ps) = " [" ++ go aliases ++ "]" ++ parameterLine ps
    parameterLine ((PI aliases _ _):ps) = " [" ++ go aliases ++ "]" ++ parameterLine ps
    parameterLine ((PS aliases _ _):ps) = " [" ++ go aliases ++ "]" ++ parameterLine ps
    parameterLine [] = ""    


    go [x] = x
    go (x:xs) = x ++ " | " ++ go xs
    go [] = ""
